/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Mike Radford, Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <errno.h>
#include "gdampinknoise.h"

/* based on "patest_pink.c"...
 *   
 *     Generate Pink Noise using Gardner method.
 *     Optimization suggested by James McCartney uses a tree
 *     to select which random value to replace.
 *         
 *       x x x x x x x x x x x x x x x x
 *        x   x   x   x   x   x   x   x
 *          x       x       x       x
 *              x               x
 *                      x
 *
 *   Tree is generated by counting leading zeros in an increasing index.
 * When the index is zero, no random number is selected.
 *                                                             
 *  Author: Phil Burk, http://www.softsynth.com
 *  Revision History:
 *        Copyleft 1999 Phil Burk - No rights reserved.
 */

static int gdam_pink_noise_read (GdamSource* this, GdamSample*   buf, int len);

static inline long int gdam_pink_noise_mrand48(GdamPinkNoise *this)
{
	return jrand48(this->random_state);
}
static inline void gdam_pink_noise_srand48(GdamPinkNoise *this, int seed)
{
	this->random_state[0] = 0x330e;
  	*((int *) &this->random_state[1]) = seed;
}

static void gdam_pink_noise_init_rows(GdamPinkNoise* this, int rows)
{
	int i, pmax;
	this->numRows = rows;

	this->pink_Index = 0;
	this->pink_IndexMask = (1<<this->numRows) - 1;

	pmax = (this->numRows + 1) * (1 << (PINK_RANDOM_BITS-1));
	this->pink_Scalar = 1.0/pmax;

	for (i = 0; i < this->numRows; ++i)
		this->pink_Rows[i] = 0;
	this->pink_RunningSum = 0;
}

static inline int gdam_pink_noise_get_seed(GdamPinkNoise *this)
{
  	return *((int *) &this->random_state[1]);
}
static void gdam_pink_noise_set_random_seed(GtkObject* object, GdamArg* arg)
{
  	gdam_pink_noise_srand48(GDAM_PINK_NOISE(object), GDAM_ARG_INT(arg));
}
static void gdam_pink_noise_get_random_seed(GtkObject* object, GdamArg* arg)
{
	GDAM_ARG_INT(arg) = gdam_pink_noise_get_seed(GDAM_PINK_NOISE(object));
}
static void gdam_pink_noise_set_rows(GtkObject* object, GdamArg* arg)
{
  	gdam_pink_noise_init_rows(GDAM_PINK_NOISE(object), GDAM_ARG_INT(arg));
}
static void gdam_pink_noise_get_rows(GtkObject* object, GdamArg* arg)
{
	GDAM_ARG_INT(arg) = GDAM_PINK_NOISE(object)->numRows;
}

static void gdam_pink_noise_init(GdamPinkNoise* this)
{
	gdam_pink_noise_srand48(this, 0xdeadbeef);
	gdam_pink_noise_init_rows(this, 30);
}

static void gdam_pink_noise_class_init(GdamPinkNoiseClass* noise_class)
{
	GtkObjectClass* object_class = (GtkObjectClass*)noise_class;
	GDAM_SOURCE_CLASS(noise_class)->read = gdam_pink_noise_read;
	gdam_arg_register(object_class,
	                  "random_seed",
			  GDAM_ARG_TYPE_INT,
			  gdam_pink_noise_set_random_seed,
			  gdam_pink_noise_get_random_seed,
			  GDAM_ARG_VITAL);
	gdam_arg_register(object_class,
	                  "rows",
			  GDAM_ARG_TYPE_INT,
			  gdam_pink_noise_set_rows,
			  gdam_pink_noise_get_rows,
			  GDAM_ARG_VITAL);
}

GtkType gdam_pink_noise_get_type()
{
	static GtkType noise_type = 0;
	if (!noise_type) {
		static const GtkTypeInfo noise_info = {
			"GdamPinkNoise",
			sizeof(GdamPinkNoise),
			sizeof(GdamPinkNoiseClass),
			(GtkClassInitFunc) gdam_pink_noise_class_init,
			(GtkObjectInitFunc) gdam_pink_noise_init,
			/* reserved_1*/ NULL,
			/* reserved_2*/ NULL,
			(GtkClassInitFunc) NULL
		};
		noise_type = gtk_type_unique(GDAM_TYPE_SOURCE,
		                                 &noise_info);
	}
	return noise_type;
}

static int gdam_pink_noise_read(GdamSource* this_source,
		GdamSample* buf, int len)
{
	GdamPinkNoise *this = GDAM_PINK_NOISE(this_source);
	int i;

	for (i = 0; i < len; ++i) {
		gint16 output_sample;
		long int newRandom;
		long int sum;
		double output;

		/* Increment and mask index. */
		++this->pink_Index;
		this->pink_Index &= this->pink_IndexMask;

		/* If index is zero, don't update any random values. */
		if (this->pink_Index) {
			/* Determine how many trailing zeros in PinkIndex. */
			/* This algorithm will hang if n==0 so test first. */
			int n, numZeros;
			for (n = this->pink_Index, numZeros = 0; (n&1) == 0;
			     n >>= 1, ++numZeros)
				;

			/* Replace the indexed ROWS random value. Subtract
			 * and add back to RunningSum instead of adding all
			 * the random values together. Only one changes each
			 * time. */

			this->pink_RunningSum -= this->pink_Rows[numZeros];
			newRandom = gdam_pink_noise_mrand48(this) >>
					PINK_RANDOM_SHIFT;
			this->pink_RunningSum += newRandom;
			this->pink_Rows[numZeros] = newRandom;
		}

		/* Add extra white noise value. */
		newRandom = gdam_pink_noise_mrand48(this) >> PINK_RANDOM_SHIFT;
		sum = this->pink_RunningSum + newRandom;

		output = this->pink_Scalar * sum;
		output *= 32768.0;
		output_sample = output;

		buf[i].s[0] = output_sample;
		buf[i].s[1] = output_sample;
	}
	return len;
}

